# coding=utf-8

import tensorflow as tf
# from tf_geometric.sparse.sparse_adj import SparseAdj
from tf_sparse import SparseMatrix

from tf_geometric.nn.conv.gcn import gcn_norm_adj
import tf_sparse as tfs


def appnp(x, edge_index, edge_weight, kernels, biases,
          dense_activation=tf.nn.relu, activation=None,
          k=10, alpha=0.1,
          dense_drop_rate=0.0, last_dense_drop_rate=0.0, edge_drop_rate=0.0,
          cache=None, training=False):

    """
    Functional API for Approximate Personalized Propagation of Neural Predictions (APPNP).

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param kernels: List[Tensor], shape of each Tensor: [num_features, num_output_features], weights
    :param biases: List[Tensor], shape of each Tensor: [num_output_features], biases
    :param dense_activation: Activation function to use for the dense layers,
        except for the last dense layer, which will not be activated.
    :param activation: Activation function to use for the output.
    :param k: Number of propagation power iterations.
    :param alpha: Teleport Probability.
    :param dense_drop_rate: Dropout rate for the output of every dense layer (except the last one).
    :param last_dense_drop_rate: Dropout rate for the output of the last dense layer.
        last_dense_drop_rate is usually set to 0.0 for classification tasks.
    :param edge_drop_rate: Dropout rate for the edges/adj used for propagation.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        To use @tf_utils.function with gcn, you should cache the noremd edge information before the first call of the gcn.

        - (1) If you're using OOP APIs tfg.layers.GCN:

              gcn_layer.build_cache_for_graph(graph)

        - (2) If you're using functional API tfg.nn.gcn:

              from tf_geometric.nn.conv.gcn import gcn_build_cache_for_graph
              gcn_build_cache_for_graph(graph)

    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tfs.shape(x)[0]
    # updated_edge_index, normed_edge_weight = gcn_norm_edge(edge_index, num_nodes, edge_weight, cache=cache)
    sparse_adj = SparseMatrix(edge_index, edge_weight, [num_nodes, num_nodes])
    normed_sparse_adj = gcn_norm_adj(sparse_adj, cache=cache)\
        .dropout(edge_drop_rate, training=training)

    num_dense_layers = len(kernels)

    h = x

    # MLP Encoder
    if kernels is not None:

        for i, (kernel, bias) in enumerate(zip(kernels, biases)):
            # SparseTensor is usually used for one-hot node features (For example, feature-less nodes.)
            if isinstance(h, tf.sparse.SparseTensor):
                h = tf.sparse.sparse_dense_matmul(h, kernel)
            else:
                h = h @ kernel

            if bias is not None:
                h += bias

            if i < num_dense_layers - 1:
                if dense_activation is not None:
                    h = dense_activation(h)
                if training and dense_drop_rate > 0.0:
                    h = tf.compat.v2.nn.dropout(h, dense_drop_rate)
            else:
                if training and last_dense_drop_rate > 0.0:
                    h = tf.compat.v2.nn.dropout(h, last_dense_drop_rate)

    output = h

    for i in range(k):
        output = normed_sparse_adj @ output
        output = output * (1.0 - alpha) + h * alpha

    if activation is not None:
        output = activation(output)

    return output




# coding=utf-8

import numpy as np
from tf_sparse import SparseMatrix
from tf_geometric.data.dataset import DownloadableDataset
import os
from scipy.io import loadmat


def _csc_to_edge_index(x):
    x = x.tocoo()
    return np.stack([x.row, x.col], axis=0)


def _csc_to_sparse_matrix(x):
    x = x.tocoo()
    index = np.stack([x.row, x.col], axis=0)
    value = x.data.astype(np.float64)
    return SparseMatrix(index, value, shape=x.shape)


class _BaseAbnormalMATDataset(DownloadableDataset):
    def __init__(self, dataset_name, dataset_root_path=None):
        super().__init__(dataset_name,
                         download_urls=["https://github.com/CrawlScript/gnn_datasets/raw/master/Abnormal/{}.zip".format(
                             dataset_name)],
                         download_file_name="{}.zip".format(dataset_name),
                         cache_name=None, dataset_root_path=dataset_root_path)

    def process(self):
        mat_path = os.path.join(self.raw_root_path, "{}.mat".format(self.dataset_name))
        data = loadmat(mat_path)

        # x = _csc_to_sparse_matrix(data["features"])
        x = data["features"].tocoo().astype(np.float64)
        y = data["label"][0].astype(np.int64)

        edge_index_dict = {}

        for key, value in data.items():
            if key.startswith("net_") or key == "homo":
                edge_index = _csc_to_edge_index(value)
                edge_index_dict[key] = edge_index

        return x, edge_index_dict, y


class FDYelpChiDataset(_BaseAbnormalMATDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("fd_yelp_chi", dataset_root_path)


class FDAmazonDataset(_BaseAbnormalMATDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("fd_amazon", dataset_root_path)



# coding=utf-8
import tensorflow as tf


def segment_op_with_pad(segment_op, data, segment_ids, num_segments):

    sort_index = tf.argsort(segment_ids)
    sorted_segment_ids = tf.gather(segment_ids, sort_index)
    sorted_data = tf.gather(data, sort_index)

    reduced_data = segment_op(sorted_data, sorted_segment_ids)
    num_paddings = num_segments - tf.shape(reduced_data)[0]

    pad_shape = tf.concat([
        [num_paddings],
        tf.shape(data)[1:]
    ], axis=0)
    pads = tf.zeros(pad_shape, dtype=reduced_data.dtype)
    outputs = tf.concat(
        [reduced_data, pads],
        axis=0
    )
    return outputs


def segment_softmax(data, segment_ids, num_segments):
    max_values = tf.math.unsorted_segment_max(data, segment_ids, num_segments=num_segments)
    gathered_max_values = tf.gather(max_values, segment_ids)
    exp = tf.exp(data - tf.stop_gradient(gathered_max_values))
    denominator = tf.math.unsorted_segment_sum(exp, segment_ids, num_segments=num_segments) + 1e-8
    gathered_denominator = tf.gather(denominator, segment_ids)
    score = exp / gathered_denominator
    return score


def segment_count(index, num_segments=None):
    data = tf.ones_like(index)
    if num_segments is None:
        num_segments = tf.reduce_max(index) + 1
    return tf.math.unsorted_segment_sum(data, index, num_segments=num_segments)



import tf_geometric as tfg

g, (train_index, valid_index, test_index) = tfg.datasets.NARSACMDataset().load_data()

print(g)
asdfasdf


def load_acm(device, args):
    g, labels, n_classes, train_nid, val_nid, test_nid = load_acm_raw()

    # my
    g = g.to(device)

    features = g.nodes["paper"].data["feat"]

    path = args.use_emb
    author_emb = torch.load(os.path.join(path, "author.pt")).float()
    field_emb = torch.load(os.path.join(path, "field.pt")).float()

    g.nodes["author"].data["feat"] = author_emb.to(device)
    g.nodes["field"].data["feat"] = field_emb.to(device)
    g.nodes["paper"].data["feat"] = features.to(device)
    paper_dim = g.nodes["paper"].data["feat"].shape[1]
    author_dim = g.nodes["author"].data["feat"].shape[1]
    assert(paper_dim >= author_dim)
    if paper_dim > author_dim:
        print(f"Randomly embedding features from dimension {author_dim} to {paper_dim}")
        author_feat = g.nodes["author"].data.pop("feat")
        field_feat = g.nodes["field"].data.pop("feat")
        rand_weight = torch.Tensor(author_dim, paper_dim).uniform_(-0.5, 0.5).to(device)
        g.nodes["author"].data["feat"] = torch.matmul(author_feat, rand_weight)
        g.nodes["field"].data["feat"] = torch.matmul(field_feat, rand_weight)

        # g.nodes["author"].data["feat"] = (torch.rand(g.nodes["author"].data["feat"].size()) - 0.5).to(device)

    # for ntype in g.ntypes:
    #     x = g.ndata["feat"][ntype]
    #     print(ntype, x.size(), x.mean())
    #     print(x[:2])
    #
    # asdfadsf

    labels = labels.to(device)
    train_nid, val_nid, test_nid = np.array(train_nid), np.array(val_nid), np.array(test_nid)

    return g, labels, n_classes, train_nid, val_nid, test_nid




import tf_geometric as tfg

# hetero_graph, train_mask_dict, test_mask_dict = HGBDataset("acm").load_data()
hetero_graph, train_mask_dict, test_mask_dict = tfg.datasets.HGBIMDBDataset().load_data()

print(hetero_graph)
print(hetero_graph.x_dict)